from __future__ import annotations

from typing import Any, NamedTuple, Sequence, cast

from sqlalchemy.ext.asyncio import AsyncConnection
from sqlalchemy.sql import FromClause, tuple_
from sqlalchemy.sql.expression import ColumnElement

from fluid import settings
from fluid.utils.errors import ValidationError

from .crud import CrudDB, Row, Select, column_value_to_python
from .cursor import Cursor, CursorEntry


class Search(NamedTuple):
    search_fields: tuple[str, ...]
    search_text: str


class Pagination(NamedTuple):
    order_by_fields: tuple[str, ...]
    limit: int
    filters: dict[str, Any]
    search: Search | None
    cursor: Cursor | None
    desc: bool = False

    @classmethod
    def create(
        cls,
        *order_by_fields: str,
        cursor: str = "",
        limit: int = 0,
        filters: dict[str, Any] | None = None,
        search: Search | None = None,
        desc: bool = False,
    ) -> Pagination:
        if limit < 0:
            raise ValidationError("limit must be greater than or equal to 0")
        if cursor:
            if limit > 0:
                raise ValidationError("limit cannot be provided with cursor")
            if filters:
                raise ValidationError("filters cannot be provided with cursor")
            if search and search.search_text:
                raise ValidationError("search text cannot be provided with cursor")
            decoded_cursor = Cursor.decode(cursor, order_by_fields)
            limit = decoded_cursor.limit
            filters = decoded_cursor.filters
            if search:
                search = search._replace(search_text=decoded_cursor.search_text)
        else:
            decoded_cursor = None
            limit = min(
                limit or settings.DEFAULT_PAGINATION_LIMIT,
                settings.DEFAULT_PAGINATION_MAX_LIMIT,
            )
        return cls(
            order_by_fields=order_by_fields,
            cursor=decoded_cursor,
            limit=limit,
            filters=filters or {},
            search=search,
            desc=desc,
        )

    @property
    def order_by_fields_sign(self) -> tuple[str, ...]:
        if self.desc:
            return tuple(f"-{field}" for field in self.order_by_fields)
        return self.order_by_fields

    async def execute(
        self,
        db: CrudDB,
        table: FromClause,
        *,
        conn: AsyncConnection | None = None,
    ) -> tuple[Sequence[Row], str]:
        sql_query = self.query(db, table)
        async with db.ensure_connection(conn) as conn:
            result = await conn.execute(sql_query)
        data = result.all()
        cursor = ""
        if len(data) > self.limit:
            cursor = self._encode_cursor(data[-1])
            data = data[:-1]
        return data, cursor

    def query(self, db: CrudDB, table: FromClause) -> Select:
        sql_query = cast(
            Select,
            db.get_query(table, table.select(), params=self.filters),
        )
        if self.search:
            sql_query = db.search_query(
                table,
                sql_query,
                self.search.search_fields,
                self.search.search_text,
            )
        start_clause = self._start_clause(table)
        if start_clause is not None:
            sql_query = sql_query.where(start_clause)
        columns = db.order_by_columns(table, self.order_by_fields_sign)
        return sql_query.order_by(*columns).limit(self.limit + 1)

    def _start_clause(self, table: FromClause) -> ColumnElement[bool] | None:
        if self.cursor:
            columns = []
            values = []
            for name, entry in zip(self.order_by_fields, self.cursor.entries):
                column = getattr(table.c, name)
                values.append(column_value_to_python(column, entry.value))
                columns.append(column)
            if self.desc:
                if len(columns) == 1:
                    return columns[0] <= values[0]
                else:
                    return tuple_(*columns) <= values
            else:
                if len(columns) == 1:
                    return columns[0] >= values[0]
                else:
                    return tuple_(*columns) >= values
        return None

    def _encode_cursor(self, row: Row) -> str:
        fields = (
            field[1:] if field.startswith("-") else field
            for field in self.order_by_fields
        )
        return Cursor(
            entries=tuple(CursorEntry(field, getattr(row, field)) for field in fields),
            limit=self.limit,
            filters=self.filters,
            search_text=self.search.search_text if self.search else "",
        ).encode()



import types
from copy import deepcopy
from typing import Any, Optional, Tuple, Type, TypeVar, Union, get_args, get_origin

from pydantic import BaseModel, create_model
from pydantic.fields import FieldInfo


def is_subclass(value: Any, cls: type) -> bool:
    """A safe version of issubclass that returns False if value is not a class."""
    try:
        return issubclass(value, cls)
    except TypeError:
        return False


def make_field_optional(field: FieldInfo, default: Any = None) -> Tuple[Any, FieldInfo]:
    """Make a field optional by wrapping it in Optional or Union"""
    new = deepcopy(field)
    if get_origin(field.annotation) is types.UnionType:
        as_none = False
        new_args = []
        for arg in get_args(field.annotation):
            if is_subclass(arg, BaseModel):
                arg = make_partial_model(arg)
            elif arg is type(None):
                as_none = True
            new_args.append(arg)
        if not as_none:
            new_args.append(type(None))
        annotation = Union[tuple(new_args)]  # type: ignore[valid-type]
    elif is_subclass(field.annotation, BaseModel):
        annotation = make_partial_model(field.annotation)  # type: ignore
    else:
        annotation = Optional[field.annotation]
    new.default = default
    new.annotation = annotation  # type: ignore[assignment]
    return (new.annotation, new)


BaseModelT = TypeVar("BaseModelT", bound=BaseModel)


def make_partial_model(model: Type[BaseModelT]) -> Type[BaseModelT]:
    return create_model(  # type: ignore
        f"Partial{model.__name__}",
        __base__=model,
        __module__=model.__module__,
        **{
            field_name: make_field_optional(field_info)
            for field_name, field_info in model.model_fields.items()
        },
    )



from .broker import TaskBroker
from .consumer import TaskConsumer, TaskManager
from .crontab import Scheduler, crontab
from .every import every
from .models import QueuedTask, Task, TaskInfo, TaskPriority, TaskRun, TaskState, task
from .scheduler import TaskScheduler

__all__ = [
    "Scheduler",
    "TaskPriority",
    "TaskState",
    "TaskScheduler",
    "task",
    "Task",
    "TaskManager",
    "TaskConsumer",
    "TaskInfo",
    "TaskRun",
    "QueuedTask",
    "TaskBroker",
    "crontab",
    "every",
]



from __future__ import annotations

import json
from abc import ABC, abstractmethod
from functools import cached_property
from typing import TYPE_CHECKING, Any, Iterable
from uuid import uuid4

from redis.asyncio import Redis
from redis.asyncio.lock import Lock
from yarl import URL

from fluid import settings
from fluid.utils.redis import FluidRedis

from .errors import UnknownTaskError
from .models import Task, TaskInfo, TaskInfoUpdate, TaskPriority, TaskRun

if TYPE_CHECKING:  # pragma: no cover
    from .consumer import TaskManager


_brokers: dict[str, type[TaskBroker]] = {}


def broker_url_from_env() -> URL:
    return URL(settings.BROKER_URL)


class TaskRegistry(dict[str, Task]):
    def periodic(self) -> Iterable[Task]:
        for task in self.values():
            yield task


class TaskBroker(ABC):
    def __init__(self, url: URL) -> None:
        self.url: URL = url
        self.registry: TaskRegistry = TaskRegistry()

    @property
    @abstractmethod
    def task_queue_names(self) -> tuple[str, ...]:
        """Names of the task queues"""

    @abstractmethod
    async def queue_task(self, task_run: TaskRun) -> None:
        """Queue a task"""

    @abstractmethod
    async def get_task_run(self, task_manager: TaskManager) -> TaskRun | None:
        """Get a Task run from the task queue"""

    @abstractmethod
    async def queue_length(self) -> dict[str, int]:
        """Length of task queues"""

    @abstractmethod
    async def get_tasks_info(self, *task_names: str) -> list[TaskInfo]:
        """List of TaskInfo objects"""

    @abstractmethod
    async def update_task(self, task: Task, params: dict[str, Any]) -> TaskInfo:
        """Update a task dynamic parameters"""

    @abstractmethod
    async def close(self) -> None:
        """Close the broker on shutdown"""

    @abstractmethod
    def lock(self, name: str, timeout: float | None = None) -> Lock:
        """Create a lock"""

    def new_uuid(self) -> str:
        return uuid4().hex

    async def filter_tasks(
        self,
        scheduled: bool | None = None,
        enabled: bool | None = None,
    ) -> list[Task]:
        task_info = await self.get_tasks_info()
        task_map = {info.name: info for info in task_info}
        tasks = []
        for task in self.registry.values():
            if scheduled is not None and bool(task.schedule) is not scheduled:
                continue
            if enabled is not None and task_map[task.name].enabled is not enabled:
                continue
            tasks.append(task)
        return tasks

    def task_from_registry(self, task: str | Task) -> Task:
        if isinstance(task, Task):
            self.register_task(task)
            return task
        else:
            if task_ := self.registry.get(task):
                return task_
            raise UnknownTaskError(task)

    def register_task(self, task: Task) -> None:
        self.registry[task.name] = task

    async def enable_task(self, task_name: str, enable: bool = True) -> TaskInfo:
        """Enable or disable a registered task"""
        task = self.registry.get(task_name)
        if not task:
            raise UnknownTaskError(task_name)
        return await self.update_task(task, dict(enabled=enable))

    @classmethod
    def from_url(cls, url: str = "") -> TaskBroker:
        p = URL(url or broker_url_from_env())
        if factory := _brokers.get(p.scheme):
            return factory(p)
        raise RuntimeError(f"Invalid broker {p}")

    @classmethod
    def register_broker(cls, name: str, factory: type[TaskBroker]) -> None:
        _brokers[name] = factory


class RedisTaskBroker(TaskBroker):
    """A simple task broker based on redis lists"""

    @cached_property
    def redis(self) -> FluidRedis:
        return FluidRedis.create(str(self.url.with_query({})), name=self.name)

    @property
    def redis_cli(self) -> Redis[bytes]:
        return self.redis.redis_cli

    @property
    def name(self) -> str:
        return self.url.query.get("name", "redis-task-broker")

    @cached_property
    def task_queue_names(self) -> tuple[str, ...]:
        priorities = self.url.query.get("queues")
        if priorities is None:
            return tuple(self.task_queue_name(p) for p in TaskPriority)
        elif priorities:
            return tuple(
                self.task_queue_name(TaskPriority[p]) for p in priorities.split(",")
            )
        else:
            return ()

    def task_hash_name(self, name: str) -> str:
        return f"{self.name}-tasks-{name}"

    def task_queue_name(self, priority: TaskPriority) -> str:
        return f"{self.name}-queue-{priority.name}"

    async def get_tasks_info(self, *task_names: str) -> list[TaskInfo]:
        pipe = self.redis_cli.pipeline()
        names = task_names or self.registry
        requested_task_names = []
        for name in names:
            if name in self.registry:
                requested_task_names.append(name)
                pipe.hgetall(self.task_hash_name(name))
        tasks_info = await pipe.execute()
        return [
            self._decode_task(self.registry[name], task_info)
            for name, task_info in zip(requested_task_names, tasks_info)
        ]

    async def update_task(self, task: Task, params: dict[str, Any]) -> TaskInfo:
        pipe = self.redis_cli.pipeline()
        info = json.loads(TaskInfoUpdate(**params).model_dump_json())
        pipe.hset(
            self.task_hash_name(task.name),
            mapping={name: json.dumps(value) for name, value in info.items()},
        )
        pipe.hgetall(self.task_hash_name(task.name))
        _, info = await pipe.execute()
        return self._decode_task(task, info)

    async def queue_length(self) -> dict[str, int]:
        if self.task_queue_names:
            pipe = self.redis_cli.pipeline()
            for name in self.task_queue_names:
                pipe.llen(name)
            result = await pipe.execute()
            return dict(zip(TaskPriority, result))
        return {}

    async def close(self) -> None:
        """Close the broker on shutdown"""
        await self.redis.close()

    async def get_task_run(self, task_manager: TaskManager) -> TaskRun | None:
        if self.task_queue_names:
            if redis_data := await self.redis_cli.brpop(
                self.task_queue_names,
                timeout=1,
            ):
                data = json.loads(redis_data[1])
                data.update(
                    task=self.task_from_registry(data["task"]),
                    task_manager=task_manager,
                )
                return TaskRun(**data)
        return None

    async def queue_task(self, task_run: TaskRun) -> None:
        await self.redis_cli.lpush(
            self.task_queue_name(task_run.priority),
            task_run.model_dump_json(),
        )

    def lock(self, name: str, timeout: float | None = None) -> Lock:
        return self.redis_cli.lock(name, timeout=timeout)

    def _decode_task(self, task: Task, data: dict[bytes, Any]) -> TaskInfo:
        info = {name.decode(): json.loads(value) for name, value in data.items()}
        return task.info(**info)


TaskBroker.register_broker("redis", RedisTaskBroker)
TaskBroker.register_broker("rediss", RedisTaskBroker)



from typing import Any, Container

from pydantic import BaseConfig, BaseModel, create_model
from pydantic.fields import FieldInfo
from sqlalchemy.inspection import inspect


class OrmConfig(BaseConfig):
    from_attributes = True


def sqlalchemy_to_pydantic(
    db_model: Any,
    model_name: str,
    *,
    config: type = OrmConfig,
    exclude: Container[str] = (),
    optional: bool = False,
) -> type[BaseModel]:
    mapper = inspect(db_model)
    if mapper is None:
        raise ValueError(f"Could not inspect {db_model}")
    fields = {}
    for column in mapper.columns:
        name = str(column.key)
        if name in exclude:
            continue
        python_type: Any = None
        if hasattr(column.type, "impl"):
            if hasattr(column.type.impl, "python_type"):
                python_type = column.type.impl.python_type
        elif hasattr(column.type, "python_type"):
            python_type = column.type.python_type
        assert python_type, f"Could not infer python_type for {column}"
        kwargs: dict = {}
        if optional:
            kwargs["default"] = None
            python_type = python_type | None
        elif column.default:
            if not column.default.is_callable:
                kwargs["default"] = column.default.arg
        fields[name] = (python_type, FieldInfo(**kwargs))
    return create_model(model_name, __config__=config, **fields)  # type: ignore



class TaskError(RuntimeError):
    pass


class UnknownTaskError(TaskError):
    pass


class DisabledTaskError(TaskError):
    pass


class TaskRunError(TaskError):
    pass


class TaskAbortedError(TaskError):
    pass


class TaskDecoratorError(TaskError):
    pass



from __future__ import annotations

import uvicorn
from fastapi import FastAPI

from fluid.utils.worker import Workers

from .service import FastapiAppWorkers


def app_workers(app: FastAPI) -> Workers:
    if workers := getattr(app.state, "workers", None):
        return workers
    else:
        workers = FastapiAppWorkers.setup(app)
        app.state.workers = workers
        return workers


def serve_app(app: FastAPI, host: str, port: int, reload: bool = False) -> None:
    uvicorn.run(app, port=port, host=host, log_level="info", reload=reload)



from __future__ import annotations

import json
import time
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from inspect import Traceback
from typing import Any, Callable, Generic, Self, Type, TypeVar

import httpx
from aiohttp import client

from fluid import settings

from .http_monitor import HttpPathFn, monitor_http_call


class ResponseError(RuntimeError):
    pass


ResponseType = Any
S = TypeVar("S", bound=client.ClientSession | httpx.AsyncClient)
R = TypeVar("R", bound=client.ClientResponse | httpx.Response)


class HttpResponse(ABC):
    @property
    @abstractmethod
    def url(self) -> str: ...

    @property
    @abstractmethod
    def status_code(self) -> int: ...

    @property
    @abstractmethod
    def method(self) -> str: ...

    @property
    @abstractmethod
    def headers(self) -> dict[str, str]: ...

    @abstractmethod
    async def json(self) -> ResponseType: ...

    @abstractmethod
    async def text(self) -> str: ...


class HttpResponseError(RuntimeError):
    def __init__(self, response: HttpResponse, data: ResponseType) -> None:
        self.response = response
        self.data = {
            "response": data,
            "request_url": response.url,
            "request_method": response.method,
            "response_status": self.status_code,
        }

    @property
    def status_code(self) -> int:
        return self.response.status_code

    def __str__(self) -> str:
        return json.dumps(self.data, indent=2)


@dataclass
class GenericHttpResponse(HttpResponse, Generic[R]):
    response: R

    @property
    def headers(self) -> dict[str, str]:
        return self.response.headers  # type: ignore


@dataclass
class AioHttpResponse(GenericHttpResponse[client.ClientResponse]):
    @property
    def url(self) -> str:
        return str(self.response.url)

    @property
    def status_code(self) -> int:
        return self.response.status

    @property
    def method(self) -> str:
        return self.response.method

    async def json(self) -> ResponseType:
        return await self.response.json()

    async def text(self) -> str:
        return await self.response.text()


@dataclass
class HttpxResponse(GenericHttpResponse[httpx.Response]):
    @property
    def url(self) -> str:
        return str(self.response.url)

    @property
    def status_code(self) -> int:
        return self.response.status_code

    @property
    def method(self) -> str:
        return self.response.request.method

    async def json(self) -> ResponseType:
        return self.response.json()

    async def text(self) -> str:
        return self.response.text


@dataclass
class HttpClient(Generic[S, R]):
    """Base class for Http clients"""

    session: S | None = None
    content_type: str = "application/json"
    session_owner: bool = False
    ResponseError: Type[HttpResponseError] = field(
        default=HttpResponseError, repr=False
    )
    ok_status: frozenset = field(default=frozenset((200, 201)), repr=False)
    default_headers: dict[str, str] = field(
        default_factory=lambda: {
            "user-agent": settings.HTTP_USER_AGENT,
        }
    )

    @abstractmethod
    def new_session(self, **kwargs: Any) -> S: ...

    @abstractmethod
    def new_response(self, response: R) -> GenericHttpResponse[R]: ...

    @abstractmethod
    async def close(self) -> None: ...

    def get_session(self) -> S:
        if not self.session:
            self.session_owner = True
            self.session = self.new_session()
        return self.session

    async def __aenter__(self) -> Self:
        return self

    async def __aexit__(
        self,
        exc_type: Type,
        exc: Type[BaseException],
        tb: Traceback,
    ) -> None:
        await self.close()

    async def get(self, url: str, **kwargs: Any) -> ResponseType:
        return await self.request("GET", url, **kwargs)

    async def patch(self, url: str, **kwargs: Any) -> ResponseType:
        return await self.request("PATCH", url, **kwargs)

    async def post(self, url: str, **kwargs: Any) -> ResponseType:
        return await self.request("POST", url, **kwargs)

    async def put(self, url: str, **kwargs: Any) -> ResponseType:
        return await self.request("PUT", url, **kwargs)

    async def delete(self, url: str, **kwargs: Any) -> ResponseType:
        return await self.request("DELETE", url, **kwargs)

    async def request(
        self,
        method: str,
        url: str,
        *,
        headers: dict | None = None,
        callback: Callable | bool | None = None,
        monitor_http: HttpPathFn | None = None,
        **kw: Any,
    ) -> ResponseType:
        session = self.get_session()
        _headers = self.get_default_headers()
        _headers.update(headers or ())
        method = method or "GET"
        start = time.monotonic()
        inner: R = await session.request(
            method,
            url,
            headers=_headers,
            **kw,
        )  # type: ignore
        response = self.new_response(inner)
        if monitor_http:
            monitor_http_call(
                response,
                time.monotonic() - start,
                sanitization_fn=monitor_http,
            )
        if callback:
            if callback is True:
                return response
            else:
                return await callback(response)
        if self.ok(response):
            data = await self.response_data(response)
        elif response.status_code == 204:
            data = {}
        else:
            await self.response_error(response)
        return data

    def ok(self, response: HttpResponse) -> bool:
        return response.status_code in self.ok_status

    def get_default_headers(self) -> dict[str, str]:
        headers = self.default_headers.copy()
        if self.content_type:
            headers["accept"] = self.content_type
        return headers

    @classmethod
    async def response_error(cls, response: HttpResponse) -> None:
        try:
            data = await cls.response_data(response)
        except Exception:
            data = {"message": await response.text()}
        raise cls.ResponseError(response, data)

    @classmethod
    async def response_data(cls, response: HttpResponse) -> ResponseType:
        if "text/csv" in response.headers["content-type"]:
            return await response.text()
        return await response.json()


class AioHttpClient(HttpClient[client.ClientSession, client.ClientResponse]):
    def new_session(self, **kwargs: Any) -> client.ClientSession:
        return client.ClientSession(**kwargs)

    def new_response(
        self, response: client.ClientResponse
    ) -> GenericHttpResponse[client.ClientResponse]:
        return AioHttpResponse(response)

    async def close(self) -> None:
        if self.session and self.session_owner:
            await self.session.close()
            self.session = None


class HttpxClient(HttpClient[httpx.AsyncClient, httpx.Response]):
    def new_session(self, **kwargs: Any) -> httpx.AsyncClient:
        return httpx.AsyncClient(**kwargs)

    def new_response(
        self, response: httpx.Response
    ) -> GenericHttpResponse[httpx.Response]:
        return HttpxResponse(response)

    async def close(self) -> None:
        if self.session and self.session_owner:
            await self.session.aclose()
            self.session = None


C = TypeVar("C", bound=HttpClient)


class HttpComponent(Generic[C]):
    def __init__(self, cli: C) -> None:
        self.cli = cli



from __future__ import annotations

from contextlib import asynccontextmanager
from dataclasses import dataclass, field
from typing import Any, AsyncIterator, Self

import sqlalchemy as sa
from sqlalchemy.engine import Engine, create_engine
from sqlalchemy.ext.asyncio import AsyncConnection, AsyncEngine, create_async_engine

from fluid import settings

from .cli import DbGroup
from .migration import Migration


@dataclass
class Database:
    """A container for tables in a database and a manager of asynchronous
    connections to a postgresql database
    """

    dsn: str
    """data source name, aka connection string"""
    echo: bool = settings.DBECHO
    pool_size: int = settings.DBPOOL_MAX_SIZE
    max_overflow: int = settings.DBPOOL_MAX_OVERFLOW
    metadata: sa.MetaData = field(default_factory=sa.MetaData)
    migration_path: str = ""
    _engine: AsyncEngine | None = None

    @classmethod
    def from_env(
        cls,
        *,
        dsn: str = settings.DATABASE,
        schema: str | None = settings.DATABASE_SCHEMA,
        **kwargs: Any,
    ) -> Self:
        """Create a new database conatiner from environment variables as defaults"""
        return cls(dsn=dsn, metadata=sa.MetaData(schema=schema), **kwargs)

    @property
    def tables(self) -> dict[str, sa.Table]:
        return self.metadata.tables

    @property
    def engine(self) -> AsyncEngine:
        """The :class:`sqlalchemy.ext.asyncio.AsyncEngine` creating connection
        and transactions"""
        if self._engine is None:
            self._engine = create_async_engine(
                self.dsn,
                echo=self.echo,
                pool_size=self.pool_size,
                max_overflow=self.max_overflow,
            )
        return self._engine

    @property
    def sync_engine(self) -> Engine:
        """The :class:`sqlalchemy.engine.Engine` for synchrouns operations"""
        return create_engine(self.dsn.replace("+asyncpg", ""))

    def cli(self, **kwargs: Any) -> DbGroup:
        """Create a new click group for database commands"""
        return DbGroup(self, **kwargs)

    @asynccontextmanager
    async def connection(self) -> AsyncIterator[AsyncConnection]:
        """Context manager for obtaining an asynchronous connection"""
        async with self.engine.connect() as conn:
            yield conn

    @asynccontextmanager
    async def ensure_connection(
        self,
        conn: AsyncConnection | None = None,
    ) -> AsyncIterator[AsyncConnection]:
        """Context manager for obtaining an asynchronous connection"""
        if conn:
            yield conn
        else:
            async with self.engine.connect() as conn:
                yield conn

    @asynccontextmanager
    async def transaction(self) -> AsyncIterator[AsyncConnection]:
        """Context manager for initializing an asynchronous database transaction"""
        async with self.engine.begin() as conn:
            yield conn

    @asynccontextmanager
    async def ensure_transaction(
        self,
        conn: AsyncConnection | None = None,
    ) -> AsyncIterator[AsyncConnection]:
        """Context manager for ensuring we a connection has initialized
        a database transaction"""
        if conn:
            if not conn.in_transaction():
                async with conn.begin():
                    yield conn
            else:
                yield conn
        else:
            async with self.transaction() as conn:
                yield conn

    async def close(self) -> None:
        """Close the asynchronous db engine if opened"""
        if self._engine is not None:
            engine, self._engine = self._engine, None
            await engine.dispose()

    async def ping(self) -> str:
        """Ping the database"""
        # TODO: we need a custom ping query
        async with self.connection() as conn:
            await conn.execute(sa.text("SELECT 1"))
        return "ok"

    def migration(self) -> Migration:
        """Create a new migration manager for this database"""
        return Migration(self)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

import click
from fastapi import FastAPI
from rich.console import Console
from rich.table import Table

from fluid.tools_fastapi import serve_app

from .endpoints import setup_fastapi

if TYPE_CHECKING:
    from .consumer import TaskManager
    from .models import TaskRun


class TaskManagerCLI(click.Group):
    def __init__(
        self, task_manager: TaskManager, *, app: FastAPI | None = None, **kwargs: Any
    ):
        kwargs.setdefault("commands", DEFAULT_COMMANDS)
        super().__init__(**kwargs)
        self.task_manager = task_manager
        self.app = app


def ctx_task_manager(ctx: click.Context) -> TaskManager:
    return ctx.parent.command.task_manager  # type: ignore


def ctx_app(ctx: click.Context) -> FastAPI | None:
    return ctx.parent.command.app  # type: ignore


@click.command()
@click.pass_context
def ls(ctx: click.Context) -> None:
    """list all tasks"""
    task_manager = ctx_task_manager(ctx)
    table = Table(title="Tasks")
    table.add_column("Name", style="cyan", no_wrap=True)
    table.add_column("Schedule", style="magenta")
    table.add_column("CPU bound", style="magenta")
    table.add_column("Description", style="green")
    for name in sorted(task_manager.registry):
        task = task_manager.registry[name]
        table.add_row(
            name,
            str(task.schedule),
            "yes" if task.cpu_bound else "no",
            task.description,
        )
    console = Console()
    console.print(table)


@click.command()
@click.pass_context
@click.argument("task")
@click.option(
    "--dry-run",
    is_flag=True,
    help="dry run (if the tasks supports it)",
    default=False,
)
def execute(ctx: click.Context, task: str, dry_run: bool) -> None:
    """execute a task"""
    task_manager = ctx_task_manager(ctx)
    run = task_manager.execute_sync(task, dry_run=dry_run)
    console = Console()
    console.print(task_run_table(run))


@click.command("serve", short_help="Start app server.")
@click.option(
    "--host",
    "-h",
    default="0.0.0.0",
    help="The interface to bind to",
    show_default=True,
)
@click.option(
    "--port",
    "-p",
    default=8080,
    help="The port to bind to",
    show_default=True,
)
@click.option(
    "--reload",
    is_flag=True,
    default=False,
    help="Enable auto-reload",
    show_default=True,
)
@click.pass_context
def serve(ctx: click.Context, host: str, port: int, reload: bool) -> None:
    """Run the service."""
    task_manager = ctx_task_manager(ctx)
    app = setup_fastapi(task_manager, app=ctx_app(ctx))
    serve_app(app, host, port, reload)


DEFAULT_COMMANDS = (ls, execute, serve)


def task_run_table(task_run: TaskRun) -> Table:
    table = Table(title="Task Run", show_header=False)
    color = "red" if task_run.state.is_failure else "green"
    table.add_column("Name", style="cyan")
    table.add_column("Description", style=color)
    table.add_row("name", task_run.task.name)
    table.add_row("description", task_run.task.description)
    table.add_row("run_id", task_run.id)
    table.add_row("state", task_run.state)
    if task_run.start:
        table.add_row("started", task_run.start.isoformat())
    if task_run.end:
        table.add_row("completed", task_run.end.isoformat())
    table.add_row("duration ms", str(task_run.duration_ms))
    return table



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Self

from redis.asyncio import BlockingConnectionPool, Redis

from fluid import settings
from fluid.utils import log

logger = log.get_logger(__name__)


@dataclass
class FluidRedis:
    redis_cli: Redis

    @classmethod
    def create(
        cls,
        url: str = "",
        name: str = settings.APP_NAME,
        max_connections: int = settings.REDIS_MAX_CONNECTIONS,
    ) -> Self:
        return cls(
            redis_cli=Redis(
                connection_pool=BlockingConnectionPool.from_url(
                    url or settings.REDIS_DEFAULT_URL,
                    max_connections=max_connections,
                    client_name=name,
                ),
            ),
        )

    def __str__(self) -> str:
        kwargs = self.redis_cli.connection_pool.connection_kwargs
        return "%s:%s" % (kwargs.get("host"), kwargs.get("port", 6379))

    async def close(self, *args: Any) -> None:
        await self.redis_cli.aclose(True)  # type: ignore[attr-defined]



import logging
import sys
from functools import partial

import aioconsole

from fluid import settings

logger = logging.getLogger(__name__)
CONSOLE_MESSAGE = """\
---
This console is running in the same asyncio event loop as the Service application.
Try: await asyncio.sleep(1, result=3)
---"""


class Console(aioconsole.AsynchronousConsole):
    def get_default_banner(self):
        cprt = (
            'Type "help", "copyright", "credits" ' 'or "license" for more information.'
        )
        return f"Python {sys.version} on {sys.platform}\n{cprt}\n{CONSOLE_MESSAGE}"


class ConsoleManager:
    aio_console = None

    def __init__(
        self,
        port: int = settings.AIO_BACKDOOR_PORT,
        host: str = "0.0.0.0",
    ) -> None:
        self.port = port
        self.host = host

    async def on_startup(self, app) -> None:
        self.aio_console = await aioconsole.start_interactive_server(
            partial(self._console, app), self.host, self.port
        )
        logger.warning("console running on port %i", self.port)

    async def on_cleanup(self, app) -> None:
        if self.aio_console:
            self.aio_console.close()
            await self.aio_console.wait_closed()
            self.aio_console = None

    def _console(self, app, streams=None) -> Console:
        return Console(locals={"app": app}, streams=streams)



"""Originally from https://github.com/coleifer/huey
"""

import re
from abc import ABC, abstractmethod
from datetime import datetime
from typing import NamedTuple, Set, Union

from zoneinfo import ZoneInfo

from fluid.utils.dates import UTC

dash_re = re.compile(r"(\d+)-(\d+)")
every_re = re.compile(r"\*\/(\d+)")
CI = Union[str, int]


class CronRun(NamedTuple):
    year: int
    month: int
    day: int
    week_day: int
    hour: int
    minute: int
    second: int = 0
    tz: ZoneInfo = UTC

    @property
    def datetime(self) -> datetime:
        return datetime(*self)


class Scheduler(ABC):
    def __str__(self) -> str:
        return f"{self.__class__.__name__}({self.info()})"

    @abstractmethod
    def info(self) -> str:
        pass

    @abstractmethod
    def __call__(
        self, timestamp: datetime, last_run: CronRun | None = None
    ) -> CronRun | None:
        """Return a CronRun if the given timestamp matches the schedule."""


class crontab(Scheduler):  # noqa N801
    """
    Convert a "crontab"-style set of parameters into a test function that will
    return True when the given datetime matches the parameters set forth in
    the crontab.
    For day-of-week, 0=Sunday and 6=Saturday.
    Acceptable inputs:
    * = every distinct value
    */n = run every "n" times, i.e. hours='*/4' == 0, 4, 8, 12, 16, 20
    m-n = run every time m..n
    m,n = run on m and n
    """

    def __init__(
        self,
        minute: CI = "*",
        hour: CI = "*",
        day: CI = "*",
        month: CI = "*",
        day_of_week: CI = "*",
        tz: ZoneInfo = UTC,
    ) -> None:
        self.tz: ZoneInfo = tz
        self._info = (
            f"minute {minute}; hour {hour}; day {day}; month {month}; "
            f"day_of_week {day_of_week}"
        )
        validation = (
            ("m", month, range(1, 13)),
            ("d", day, range(1, 32)),
            ("w", day_of_week, range(8)),  # 0-6, but also 7 for Sunday.
            ("H", hour, range(24)),
            ("M", minute, range(60)),
        )
        cron_settings = []

        for date_str, value, acceptable in validation:
            settings: Set[int] = set()

            if isinstance(value, int):
                value = str(value)

            for piece in value.split(","):
                if piece == "*":
                    settings.update(acceptable)
                    continue

                if piece.isdigit():
                    digit = int(piece)
                    if digit not in acceptable:
                        raise ValueError("%d is not a valid input" % digit)
                    elif date_str == "w":
                        digit %= 7
                    settings.add(digit)

                else:
                    dash_match = dash_re.match(piece)
                    if dash_match:
                        lhs, rhs = map(int, dash_match.groups())
                        if lhs not in acceptable or rhs not in acceptable:
                            raise ValueError("%s is not a valid input" % piece)
                        elif date_str == "w":
                            lhs %= 7
                            rhs %= 7
                        settings.update(range(lhs, rhs + 1))
                        continue

                    # Handle stuff like */3, */6.
                    every_match = every_re.match(piece)
                    if every_match:
                        if date_str == "w":
                            raise ValueError(
                                "Cannot perform this kind of matching"
                                " on day-of-week."
                            )
                        interval = int(every_match.groups()[0])
                        settings.update(acceptable[::interval])

            cron_settings.append(sorted(list(settings)))
        self.cron_settings = tuple(cron_settings)

    def info(self) -> str:
        return self._info

    def __call__(
        self, timestamp: datetime, last_run: CronRun | None = None
    ) -> CronRun | None:
        # assume datetime is always utc
        if timestamp.tzinfo is None:
            raise ValueError("timestamp tick must have a timezone")
        if timestamp.tzinfo != self.tz:
            # convert timestamp into cron_settings timezone
            timestamp = timestamp.astimezone(self.tz)
        year, month, day, hour, minute, _, w, _, _ = timestamp.timetuple()
        run = CronRun(year, month, day, (w + 1) % 7, hour, minute)
        if last_run == run:
            return None

        for date_piece, selection in zip(run[1:], self.cron_settings):
            if date_piece not in selection:
                return None
        return run



from typing import Annotated, Any, cast

from fastapi import APIRouter, Depends, FastAPI, HTTPException, Path, Request
from pydantic import BaseModel, Field

from fluid.scheduler import (
    TaskInfo,
    TaskManager,
    TaskPriority,
    TaskRun,
)
from fluid.scheduler.errors import UnknownTaskError
from fluid.tools_fastapi import app_workers
from fluid.utils.worker import Worker

router = APIRouter()


def get_task_manger_from_request(request: Request) -> TaskManager:
    return get_task_manger(request.app)


def get_task_manger(app: FastAPI) -> TaskManager:
    return cast(TaskManager, app.state.task_manager)


TaskManagerDep = Annotated[TaskManager, Depends(get_task_manger_from_request)]


class TaskUpdate(BaseModel):
    enabled: bool = Field(description="Task enabled or disabled")


class TaskCreate(BaseModel):
    name: str = Field(description="Task name")
    params: dict[str, Any] = Field(description="Task parameters", default_factory=dict)
    priority: TaskPriority | None = Field(default=None, description="Task priority")


@router.get(
    "/tasks",
    response_model=list[TaskInfo],
    summary="List Tasks",
    description="Retrieve a list of tasks runs",
)
async def get_tasks(task_manager: TaskManagerDep) -> list[TaskInfo]:
    return await task_manager.broker.get_tasks_info()


@router.get(
    "/tasks/status",
    response_model=dict,
    summary="Task consumer status",
    description="Retrieve a list of tasks runs",
)
async def get_task_status(task_manager: TaskManagerDep) -> dict:
    if isinstance(task_manager, Worker):
        return await task_manager.status()
    return {}


@router.post(
    "/tasks",
    response_model=TaskRun,
    summary="Queue a new Tasks",
    description="Queue a new task to be run",
)
async def queue_task(
    task_manager: TaskManagerDep,
    task: TaskCreate,
) -> TaskRun:
    try:
        return await task_manager.queue(task.name, task.priority, **task.params)
    except UnknownTaskError as exc:
        raise HTTPException(status_code=404, detail="Task not found") from exc


@router.patch(
    "/tasks/{task_name}",
    response_model=TaskInfo,
    summary="Update a task",
    description="Update a task configuration and enable/disable it",
)
async def patch_task(
    task_manager: TaskManagerDep,
    task_update: TaskUpdate,
    task_name: str = Path(title="Task name"),
) -> TaskInfo:
    try:
        return await task_manager.broker.enable_task(task_name, task_update.enabled)
    except UnknownTaskError as exc:
        raise HTTPException(status_code=404, detail="Task not found") from exc


def setup_fastapi(
    task_manager: TaskManager,
    *,
    app: FastAPI | None = None,
    **kwargs: Any,
) -> FastAPI:
    """Setup the FastAPI app"""
    app = app or FastAPI(**kwargs)
    app.include_router(router, tags=["Tasks"])
    app.state.task_manager = task_manager
    if isinstance(task_manager, Worker):
        app_workers(app).add_workers(task_manager)
    return app



from typing import Any


class FluidError(Exception):
    """Base class for all Fluid Trading errors."""


class FluidValueError(ValueError, FluidError):
    """Quant Value Error"""


class ValidationError(FluidValueError):
    """Validation Error"""

    def __init__(
        self, field: str, msg: str = "", field_type: str = "string", value: Any = None
    ):
        self.field = field
        self.field_type = field_type
        self.msg = msg or "validation error"
        self.value = value

    def __str__(self) -> str:
        return self.msg



from datetime import date, datetime, timezone
from typing import Any

from zoneinfo import ZoneInfo

UTC = ZoneInfo("UTC")


def utcnow() -> datetime:
    return datetime.now(tz=UTC)


def as_utc(dt: date | None) -> datetime:
    if dt is None:
        return utcnow()
    elif isinstance(dt, datetime):
        return dt.replace(tzinfo=UTC)
    else:
        return datetime(dt.year, dt.month, dt.day, tzinfo=UTC)


def isoformat(dt: datetime, **kwargs: Any) -> str:
    if dt.tzinfo is timezone.utc:
        return dt.replace(tzinfo=None).isoformat(**kwargs) + "Z"
    else:
        return dt.isoformat(**kwargs)



from __future__ import annotations

import base64
import json
from typing import Any, NamedTuple, Self

from fluid.utils.errors import ValidationError
from fluid.utils.text import to_bytes


class CursorEntry(NamedTuple):
    field: str
    value: Any


class Cursor(NamedTuple):
    entries: tuple[CursorEntry, ...]
    limit: int
    filters: dict[str, Any]
    search_text: str

    @classmethod
    def decode(cls, cursor: str, field_names: tuple[str, ...]) -> Self:
        try:
            base64_bytes = cursor.encode("ascii")
            cursor_bytes = base64.b64decode(base64_bytes)
            values, limit, filters, search_text = json.loads(cursor_bytes)
            if len(values) == len(field_names):
                return cls(
                    entries=tuple(map(CursorEntry, field_names, values)),
                    limit=limit,
                    filters=filters,
                    search_text=search_text,
                )
            raise ValueError
        except Exception as e:
            raise ValidationError("invalid cursor") from e

    def encode(self) -> str:
        data = tuple(e.value for e in self.entries)
        cursor_bytes = to_bytes(
            json.dumps((data, self.limit, self.filters, self.search_text))
        )
        base64_bytes = base64.b64encode(cursor_bytes)
        return base64_bytes.decode("ascii")
